/*****************************************
 * Read QSee/Zmodo cameras               *
 * Forward stream to FIFO pipe           *
 * Author: Daniel Osborne                *
 * Based on IP Cam Viewer by Robert Chou *
 * License: Public Domain                *
 *****************************************/

/* Version history
 * 0.3 - 2012-01-26
 *       Add support for DVR-8104UV/8114HV and CnM Classic 4 Cam DVR
 * 0.2 - 2011-11-12
 *       Got media port support working (for some models at least).
 *       Changed fork behavior to fix a bug, now parent only spawns children, it doesn't stream.
 * 0.1 - 2011-08-26
 *       Initial version, working mobile port support, but buggy
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/wait.h>

typedef enum bool {false=0, true=1,} bool;

// Structure for logging into ZModo Control Port
// Must be in network byte order
struct ZModoControlLogin
{
	char head[4];		//  4 header
	char user[50];		// 50 username field
	char pass[50];		// 50 password field
};	// Total size:		  104 bytes

// Structure for media port request command
struct ZModoMediaStart
{
	char head[4];		//  4 header
	char cam;		//  1 cam
	char chan;		//  1 chan
	char magic[6];		//  6 magic
	char host[30];		// 30 host
	char user[50];		// 50 username field
	char pass[50];		// 50 password field
	char foot[4];		//  4 footer
};	// Total size:		  146 bytes

#define MAX_CHANNELS 16		// maximum channels to support (I've only seen max of 16).

struct globalArgs_t {
	bool verbose;			// -v duh
	char *pipeName;			// -n name to use for filename (ch # will be appended)
	bool channel[MAX_CHANNELS];     // -c (support up to 16)
	char *hostname;			// -s hostname to connect to
	unsigned short cport;		// -p port number
	unsigned short port;		// set by control
	char *username;			// -u login username
	char *password;			// -a login password
	int timer;			// -t alarm timer
	char key[4];			// set by channel control
} globalArgs = {0};

extern char *optarg;
const char *optString = "vn:c:p:s:u:a:t:h?";
int g_childPids[MAX_CHANNELS] = {0};
int g_cleanUp = false;
char g_errBuf[256];	// This will contain the error message for perror calls

void sigHandler(int sig);
void display_usage(char *name);

int Connect(char *hostname, unsigned short port);
int ConnectMaster(int sockFd);
int ConnectControl(int sockFd, int channel);
int ConnectMedia(int sockFd, int channel);

void printBuffer(char *pbuf, size_t len)
{
		int n;

		for(n=0; n < len ; n++)
		    printf("%02x",(unsigned char)(pbuf[n]));

		for(n=0; n < len ; n++)
		    printf("%c",(unsigned char)(pbuf[n]));
		printf("\n");
}

int main(int argc, char**argv)
{
	char pipename[256];
	struct addrinfo hints, *server;
	int retval = 0;
	char recvBuf[2048];
	struct sigaction sapipe, oldsapipe, saterm, oldsaterm, saint, oldsaint, sahup, oldsahup;
	char opt;
	int loopIdx;
    
	int outPipe = -1;
	int sockFd = -1;
	int cSockFd = -1;
    int mSockFd = -1;
    
	int processCh = -1;	// Channel this process will be in charge of (-1 means parent)
    unsigned int mPort = -1;

	// Process arguments
	// Clear and set defaults
	memset(&globalArgs, 0, sizeof(globalArgs));
	
	globalArgs.hostname = "172.16.187.136";
	globalArgs.pipeName = "zmodo";
	globalArgs.username = "Admin";
	globalArgs.password = "111111";
	globalArgs.cport = 5050;

	// Read command-line
	while( (opt = getopt(argc, argv, optString)) != -1 )
	{
		switch( opt )
		{
		case 'v':
			globalArgs.verbose = true;
			break;
		case 'c':
		{
			globalArgs.channel[atoi(optarg) - 1] = true;
			break;
		}
		case 'n':
			globalArgs.pipeName = optarg;
			break;
		case 's':
			globalArgs.hostname = optarg;
			break;
		case 'p':
			globalArgs.cport = atoi(optarg);
			break;
		case 'u':
			globalArgs.username = optarg;
			break;
		case 'a':
			globalArgs.password = optarg;
			break;
		case 't':
			globalArgs.timer = atoi(optarg);
			break;
		case 'h':
			// Fall through
		case '?':
			// Fall through
		default:
			display_usage(argv[0]);
			return 0;
		}
	}

	memset(&saint, 0, sizeof(saint));
	memset(&saterm, 0, sizeof(saterm));
	memset(&sahup, 0, sizeof(sahup));
	
	// Ignore SIGPIPE
	sapipe.sa_handler = sigHandler;
	sigaction(SIGPIPE, &sapipe, &oldsapipe);

	// Handle SIGTERM & SIGING
	saterm.sa_handler = sigHandler;
	sigaction(SIGTERM, &saterm, &oldsaterm);
	saint.sa_handler = sigHandler;
	sigaction(SIGINT, &saint, &oldsaint);
	
	// SIGUSR1 is used to reset the pipe and connection
	sahup.sa_handler = sigHandler;
	sigaction(SIGUSR1, &sahup, &oldsahup);
	
	// SIGUSR2 is used to reset the pipe and connection
	sahup.sa_handler = sigHandler;
	sigaction(SIGUSR2, &sahup, &oldsahup);

    // Set up control connection
    if((sockFd = Connect(globalArgs.hostname, globalArgs.cport)) != -1 && (mPort = ConnectMaster(sockFd)) == -1) {
		printf("Error connecting to control\n");
		return 1;
    }
    close(sockFd);
    // We have the media port in mPort

	// Create a fork for each camera channel to stream
	for( loopIdx=0;loopIdx<MAX_CHANNELS;loopIdx++ )
	{
		if( globalArgs.channel[loopIdx] == true )
		{
			g_childPids[loopIdx] = fork();
			
			// Child Process
			if( g_childPids[loopIdx] == 0 )
			{
				memset(g_childPids, 0, sizeof(g_childPids));
				processCh = loopIdx;
				break;
			}
			// Error
			else if( g_childPids[loopIdx] == -1 )
			{
				printf("fork failed\n");
				return 1;
			}
		}
	}
	if( processCh == -1 )
	{
		int status = 0;
		int pid;

		signal( SIGUSR1, SIG_IGN );		// Ignore SIGUSR1 in parent process
        
		// parent process just waits for children to exit
		while( (pid = wait(&status)) > 0  && g_cleanUp != true )
		{
			if( globalArgs.verbose )
				printf("Child %i returned: %i\n", pid, status);

			if( g_cleanUp == 2 )
				g_cleanUp = false;	// Ignore SIGHUP
		}
	}
	else
	{
		// At this point, processCh contains the camera number to use
		sprintf(pipename, "/tmp/%s%i", globalArgs.pipeName, processCh);
		retval = mkfifo(pipename, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
        if(retval == 0 || (retval == -1 && errno == EEXIST)) {
            // the stream sometimes goes grey,
            // this alarm should periodically reset the stream
            if(globalArgs.timer)
                alarm(globalArgs.timer);

            while( !g_cleanUp ) {
                // Open the pipe if it wasn't previously opened, this should block until a reader is available
                if( globalArgs.verbose )
                    printf("\nCh %i: %s\n", processCh+1, "Opening pipe and connecting");
                
                // reader available, open the media stream
                if((outPipe = open(pipename, O_WRONLY)) != -1 && (cSockFd = Connect(globalArgs.hostname, globalArgs.cport)) != -1 && (mSockFd = Connect(globalArgs.hostname, mPort)) != -1) {
                    if(ConnectMedia(mSockFd, processCh) && ConnectControl(cSockFd, processCh)) {
                        signal( SIGPIPE, SIG_IGN );		// Ignore SIGPIPE

                        // Now we are connected and awaiting stream
                        while( !g_cleanUp ) {
                            int read;
                            // Read actual h264 data from camera
                            if( globalArgs.verbose ) {
                                printf("%i", processCh + 1);
                                fflush(stdout);
                            }
                            read  = recv(mSockFd, recvBuf, sizeof(recvBuf), 0);
                            
                            // Server disconnected, close the socket so we can try to reconnect
                            if( read <= 0 ) {
                                if( globalArgs.verbose )
                                    printf("\nCh %i: Socket closed. Receive result: %i\n", processCh+1, read);
                                close(outPipe);
                                close(cSockFd);
                                close(mSockFd);
                                break;
                            }
                            
                            if( globalArgs.verbose ) {
                                printf(".");
                                fflush(stdout);
                            }
                            
                            // send to pipe
                            if( (retval = write(outPipe, recvBuf, read)) == -1) {
                                if( globalArgs.verbose ) {
                                    // reader closed the pipe, wait for it to be opened again.
                                    sprintf(g_errBuf, "\nCh %i: %s", processCh+1, "Pipe closed");
                                    perror(g_errBuf);
                                }
                                close(outPipe);
                                close(cSockFd);
                                close(mSockFd);
                                break;
                            } else if(globalArgs.verbose) {
                                printf("*");
                                fflush(stdout);
                            }
                        }
                        
                        if( globalArgs.verbose )
                            printf("\nCh %i: Exited pipe writing\n", processCh+1);
                        
                        // If we receive a SIGUSR1, close and reset everything
                        // Then start the loop over again.
                        if( g_cleanUp >= 2 )
                            g_cleanUp = false;
                    }
                }
                
            }
            // Received signal to exit, cleanup
            close(outPipe);
            close(sockFd);
            close(cSockFd);

            unlink(pipename);
        }
    }
    

	// Restore old signal handler
	sigaction(SIGPIPE, &oldsapipe, NULL);
	sigaction(SIGTERM, &oldsaterm, NULL);
	sigaction(SIGINT, &oldsaint, NULL);
	sigaction(SIGUSR1, &oldsahup, NULL);
	freeaddrinfo(server);

	// Kill all children (if any)
	for( loopIdx=0;loopIdx<MAX_CHANNELS;loopIdx++ )
	{
		if( globalArgs.channel[loopIdx] > 0 )
			kill( globalArgs.channel[loopIdx], SIGTERM );
	}
	return 0;
}

int Connect(char *hostname, unsigned short port)
{
	int sockFd = -1;
	int retval;
	struct sockaddr_in serverAddr;
	struct addrinfo hints, *server;
	struct timeval tv;
	struct linger lngr;
    char* flag;
    
	lngr.l_onoff = false;
	lngr.l_linger = 0;
    
    tv.tv_sec = 20000;		// Wait 5 seconds for socket data
    tv.tv_usec = 0;
    
    memset(&serverAddr, 0, sizeof(serverAddr));
	serverAddr.sin_family = AF_INET;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_protocol = IPPROTO_TCP;
    
	if( (retval = getaddrinfo(hostname, NULL, &hints, &server)) != 0 ) {
		printf("getaddrinfo failed: %s\n", gai_strerror(retval));
	} else {
        serverAddr.sin_addr = ((struct sockaddr_in*)server->ai_addr)->sin_addr;
        serverAddr.sin_port = htons(port);
        
        sockFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        
        if(setsockopt(sockFd, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, sizeof(tv))) {
            sprintf(g_errBuf, "Control: %s", "Failed to set socket timeout");
            perror(g_errBuf);
        } else if( setsockopt(sockFd, IPPROTO_TCP, TCP_NODELAY, (char*)&flag, sizeof(flag))) {
            sprintf(g_errBuf, "Control: %s", "Failed to set TCP_NODELAY");
            perror(g_errBuf);
        } else if( setsockopt(sockFd, SOL_SOCKET, SO_LINGER, (char*)&lngr, sizeof(lngr))) {
            sprintf(g_errBuf, "Control: %s", "Failed to set SO_LINGER");
            perror(g_errBuf);
        } else if((retval = connect(sockFd, (struct sockaddr*)&serverAddr, sizeof(serverAddr))) == -1 && errno != EINPROGRESS) {
            if( globalArgs.verbose ) {
                sprintf(g_errBuf, "Control: %s", "Failed to connect");
                perror(g_errBuf);
            }
            close(sockFd);
            sockFd = -1;
        } else if( globalArgs.verbose ) {
                printf("Control: Connect result: %i\n", retval);
        }
    }
	return sockFd;
}

void display_usage(char *name)
{
	printf("Usage: %s [options]\n\n", name);
	printf(	"Where [options] is one of:\n\n"
		"    -s <string>\tIP to connect to\n"
		"    -t <int>\tSend a timer interrupt every x seconds.\n"
		"    -p <int>\tPort number to connect to\n"
		"    -c <int>\tChannels to stream (can be specified multiple times)\n"
		"    -n <string>\tBase filename of pipe (ch# will be appended)\n"
		"    -v\t\tVerbose output\n"
		"    -u <string>\tUsername\n"
		"    -a <string>\tPassword\n"
	"\n");
}

void sigHandler(int sig)
{
	switch( sig )
	{
	case SIGTERM:
	case SIGINT:
		// Kill the main loop
		g_cleanUp = true;
		break;
	case SIGUSR1:
	case SIGALRM:
		g_cleanUp = 2;
		break;
	case SIGUSR2:
		g_cleanUp = 3;
		break;
	}
}

// Takes connected socket. Sends login to control and query for media port
// returns media port or -1 for failure
int ConnectMaster(int sockFd)
{
	struct ZModoControlLogin loginBuf;
	char command[4];
    unsigned short mPort = -1, byte1, byte2;
	int retval;
	char recvBuf[8];
    
	memset(&loginBuf, 0, sizeof(loginBuf));
	loginBuf.head[0] = 0;
	loginBuf.head[1] = 0x18;
	loginBuf.head[2] = 0x64;
	loginBuf.head[3] = 0;
	strcpy(loginBuf.user, globalArgs.username);
	strcpy(loginBuf.pass, globalArgs.password);

	if( globalArgs.verbose )
	{
		printBuffer((char*)&loginBuf, sizeof(loginBuf));
	}

	retval = send(sockFd, (char*)(&loginBuf), sizeof(loginBuf), 0);

	if( globalArgs.verbose )
		printf("Control: Send result: %i\n", retval);

	// Get header length (8 bytes)
	if( retval = recv(sockFd, &recvBuf, 8, 0) == 5 )
	{
        memset(command, 0, 4);
        command[1] = 0x32;
        retval = send(sockFd, command, 4, 0);
        if( globalArgs.verbose )
            printf("Control: Send result: %i\n", retval);
        
        if( retval = recv(sockFd, &recvBuf, 7, 0) == 7 )
        {
            
            byte1 = 0x00FF & recvBuf[5];
            byte2 = 0x00FF & recvBuf[6];
            mPort = (byte2 << 8) + byte1;
            if( globalArgs.verbose )
                printf("Control: Media Port: %u\n", mPort);
        } else {
            printf("Receive 2 failed: %i\n", retval);
        }
	} else {
		printf("Receive 1 failed: %i\n", retval);
    }
    
    close(sockFd);
	return mPort;
}

int ConnectControl(int sockFd, int channel)
{
	struct ZModoMediaStart loginBuf;
	bool beenHere = false;
	int retval;
	char recvBuf[11];

	memset(&loginBuf, 0, sizeof(loginBuf));
	loginBuf.head[0] = 0;
	loginBuf.head[1] = 0x05;
	loginBuf.head[2] = 0x8e;
	loginBuf.head[3] = 0;
	loginBuf.cam = (char)channel;
	loginBuf.chan = (char)(channel + 0xa2);
	loginBuf.magic[0] = 0x17;
	loginBuf.magic[1] = 0;
	loginBuf.magic[2] = 0;
	loginBuf.magic[3] = 0x01;
	loginBuf.magic[4] = 0;
	loginBuf.magic[5] = 0;
	char foot[4];		//  4 footer
	strcpy(loginBuf.host, globalArgs.hostname);
	strcpy(loginBuf.user, globalArgs.username);
	strcpy(loginBuf.pass, globalArgs.password);
	loginBuf.foot[0] = globalArgs.key[0];
	loginBuf.foot[1] = globalArgs.key[1];
	loginBuf.foot[2] = globalArgs.key[2];
	loginBuf.foot[3] = globalArgs.key[3];

	if( globalArgs.verbose && beenHere == false )
	{
		printBuffer((char*)&loginBuf, sizeof(loginBuf));
		beenHere = true;
	}

	retval = send(sockFd, (char*)(&loginBuf), sizeof(loginBuf), 0);

	if( globalArgs.verbose )
		printf("Ch %i: Send result, %i\n", channel+1, retval);

	// Get response length (11 bytes)
	retval = recv(sockFd, &recvBuf, 11, 0);
	if( retval != 11 )
	{
		printf("Ch %i: Login reponse failed, %i\n", channel + 1, retval);
		return 0;
	} else if(globalArgs.verbose){
		printf("Ch %i: Control ready\n", channel + 1);
	}
	// If we got here, we have the media port
	return 1;
}
// This is less compatible, but more reliable than the mobile mode
// Output is 704x480@25fps 1200kbit/s VBR
int ConnectMedia(int sockFd, int channel)
{
	int retval;
	char recvBuf[12];

	// Get response length (12 bytes)
	retval = recv(sockFd, &recvBuf, 12, 0);
	if( retval != 12 )
	{
		printf("Ch %i: Media ready failed, %i\n", channel + 1, retval);
		return 0;
	} else if(globalArgs.verbose){
		printf("Ch %i: Media ready\n", channel + 1);
	}
	
	globalArgs.key[0] = recvBuf[8];
	globalArgs.key[1] = recvBuf[9];
	globalArgs.key[2] = recvBuf[10];
	globalArgs.key[3] = recvBuf[11];
	// If we got here, we have the media port
	return 1;
}
